/*
 * Copyright 2009 Red Hat, Inc.
 * Red Hat licenses this file to you under the Apache License, version
 * 2.0 (the "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied.  See the License for the specific language governing
 * permissions and limitations under the License.
 */
package org.hornetq.jms.example;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

import java.util.ArrayList;
import java.util.List;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import javax.management.MBeanServerConnection;
import javax.management.MBeanServerInvocationHandler;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageProducer;
import javax.jms.Queue;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.naming.InitialContext;

import org.hornetq.common.example.HornetQExample;
import org.hornetq.api.core.management.ObjectNameBuilder;
import org.hornetq.api.jms.management.JMSQueueControl;

/**
 * A simple JMS example that sends and consume message transactionally.
 *
 * @author <a href="hgao@redhat.com">Howard Gao</a>
 */
public class TransactionalExample extends HornetQExample
{
   static final long ONE_SECOND = 1000l;
   static final long TEST_TIME = 1200000l;
   static final int MAX_MESSAGES = 20;
   static final int CONNS = 11;
   static final long RECEIVE_TIMEOUT = 500l;
   static final int THREADS_CONSUMER = 4;
   static final int THREADS_PRODUCER = 4;
   /* 
    * I originally wrote this program to 
    * closely represent the real world 
    * program where we see these errors.
    * This includes connection pooling
    * and creating new sessions each time
    * there is new work to do. However, after
    * reading the bug reporting instructions
    * and re-reading the performance tuning
    * section, I built some simple flags to
    * test with and without connection and
    * session re-use. The error condition 
    * seems to manifest itself more quickly
    * with connection and session re-use.
    *
    * These flags were implemented separately,
    * but it doesn't really make sense to 
    * set SINGLE_CONN = false and also set
    * REUSE_SESSION = true.
    * 
    * TL;DR - Set SINGLE_CONN and REUSE_SESSION
    * to the same value.
    */
   static final boolean SINGLE_CONN = true;
   static final boolean REUSE_SESSION = true;

   public static void main(final String[] args)
   {
      new TransactionalExample().run(args);
   }

   static class TestConnectionPool {
      private ConnectionFactory cf;

      private List<Connection> allConnections;
      private HashSet<Connection> leasedConnections;
      private List<Connection> unusedConnections;
      private boolean shutdown;

      public TestConnectionPool(ConnectionFactory cf) {
         this.cf = cf;
         this.allConnections = new ArrayList<Connection>(CONNS);
         this.leasedConnections = new HashSet<Connection>();
         this.unusedConnections = new ArrayList<Connection>(CONNS);
         this.shutdown = false;
      }

      public synchronized Connection borrowConnection() {
         if(shutdown) {
            throw new IllegalStateException("TestConnectionPool already shut down.");
         }
         if(SINGLE_CONN) {
            if(allConnections.isEmpty()) {
               try {
                  System.out.println("Creating new connection.");
                  Connection connection = cf.createConnection();
                  connection.start();

                  allConnections.add(connection);

               } catch(JMSException jmse) {
                  throw new RuntimeException(jmse);
               }
            }
            return allConnections.get(0);
         }

         if(unusedConnections.isEmpty()) {
            if(allConnections.size() >= CONNS) {
               throw new IllegalStateException("Unable to create more connections.");
            }
            try {
               System.out.println("Creating new connection.");
               Connection connection = cf.createConnection();
               connection.start();

               allConnections.add(connection);
               leasedConnections.add(connection);
               return connection;

            } catch(JMSException jmse) {
               throw new RuntimeException(jmse);
            }
         } else {
            Connection connection = unusedConnections.remove(0);
            leasedConnections.add(connection);
            return connection;
         }
      }

      public synchronized void returnConnection(Connection connection) {
         if(SINGLE_CONN) {
            // Do nothing.
            return;
         }

         final boolean removed = leasedConnections.remove(connection);
         assert(removed);
         unusedConnections.add(connection);
      }

      public synchronized void shutdown() {
         this.shutdown = true;
         for(Connection connection : allConnections) {
            try {
               connection.close();
            } catch(JMSException jmse) {
               jmse.printStackTrace();
            }
         }
         allConnections.clear();
         leasedConnections.clear();
         unusedConnections.clear();
      }
   }

   static class JmsTester {
      private TestConnectionPool pool;
      private Queue queue;
      private Session session;

      public void setQueue(Queue queue) {
         this.queue = queue;
      }
      public Queue getQueue() {
         return queue;
      }
      public void setConnectionPool(TestConnectionPool pool) {
         this.pool = pool;
      }
      public Connection getConnection() {
         return pool.borrowConnection();
      }
      public void returnConnection(Connection conn) {
         pool.returnConnection(conn);
      }
      public Session getSession(Connection conn) throws JMSException {
         if(!REUSE_SESSION) {
            //System.out.println("Creating new session");
            return conn.createSession(true, Session.SESSION_TRANSACTED);
         }
         if(session == null) {
            System.out.println("Creating new session");
            session = conn.createSession(true, Session.SESSION_TRANSACTED);
         }
         return session;
      }
 
      protected Random rand = new Random();

      private boolean enabled = true;

      public boolean isEnabled() {
          return enabled;
      }

      public void setEnabled(boolean enabled) {
          this.enabled = enabled;
      }

      private Map<String, Object> getControlProps() {
         final Map<String, Object> environment = new HashMap<String, Object>();
         environment.put(
            JMXConnector.CREDENTIALS,
            new String[] { "controlRole", "password" });

         return environment;
      }

      long countMessages() {
         try {
            final Map<String, Object> environment = getControlProps();
            final JMXServiceURL serviceURL = new JMXServiceURL("service:jmx:rmi:///jndi/rmi://localhost:3000/jmxrmi");
            final JMXConnector connector = JMXConnectorFactory.connect(serviceURL, environment);
            final MBeanServerConnection mbsc = connector.getMBeanServerConnection();
            final ObjectName on = ObjectNameBuilder.DEFAULT.getJMSQueueObjectName("exampleQueue");

            final JMSQueueControl qControl = MBeanServerInvocationHandler.newProxyInstance(mbsc, on, JMSQueueControl.class, false);

            final long count = qControl.countMessages(null);

            connector.close();
            return count;
         } catch(Exception e) {
            throw new RuntimeException(e);
         }
      }

      void sleepRand() {
          final int sleepTime = rand.nextInt(500);
          try {
             Thread.sleep(sleepTime);
          } catch(final InterruptedException ie) {}
      }
   }

   static class TestMessageProducer extends JmsTester implements Runnable {

      private static int counter = 0;

      synchronized static int count() {
         return counter++;
      }

      synchronized static int getCount() {
         return counter;
      }

      public void run() {
         while(isEnabled()) {
            Connection conn = getConnection();
            Session session = null;
            try {
               session = getSession(conn);
               final MessageProducer producer = session.createProducer(getQueue());

               final int messagesToSend = rand.nextInt(MAX_MESSAGES);
               System.out.println("+ " + messagesToSend + " messages");
               for(int i=0; i < messagesToSend; i++) {
                  final Message m = session.createTextMessage("This is a text message");
                  producer.send(m);
                  count();
               }

               session.commit();
               producer.close();
            }
            catch(final Exception e) {
               if(session != null) {
                  try {
                     session.rollback();
                  } catch(Exception e2) {}
               }
               setEnabled(false);
               e.printStackTrace();
               throw new RuntimeException(e);
            }
            finally {
               try {
                  if(session != null && !REUSE_SESSION) {
                     if(session.getTransacted()) {
                        session.commit();
                     }
                     session.close();
                  }
               } catch(Exception e) {
                  e.printStackTrace();
                  throw new RuntimeException(e);
               } finally {
                  returnConnection(conn);
               }
            }

            sleepRand();
         }
      }
   }

   static class TestMessageConsumer extends JmsTester implements Runnable {

      private static int counter = 0;

      synchronized static int count() {
         return counter++;
      }

      synchronized static int getCount() {
         return counter;
      }

      private TransactionalExample example;
      public void setExample(TransactionalExample example) {
         this.example = example;
      }

      public void run() {
         int countFailures = 0;
         while(isEnabled()) {
            Session session = null;
            final int messagesToRead = rand.nextInt(MAX_MESSAGES);
            System.out.println("- " + messagesToRead + " messages");

            try {
               final long count = countMessages();
               if(count < messagesToRead) {
                  System.out.println("* Aborted read. Count was only " + count);
                  sleepRand();
                  continue;
               }
            } catch(final Exception e) {
               e.printStackTrace();
               setEnabled(false);
               break;
            }
            Connection conn = getConnection();

            try {
               session = getSession(conn);

               final MessageConsumer consumer = session.createConsumer(getQueue());
               int messagesRead = 0;
               for(int i=0; i < messagesToRead; i++) {
                  final Message m = consumer.receive(RECEIVE_TIMEOUT);
                  if(m != null) {
                     messagesRead++;
                     count();
                  }
               }
               if(messagesRead != messagesToRead) {
                  System.err.println("*** EXPECTED " + messagesToRead + " BUT FOUND " + messagesRead + " ***" );
                  countFailures++;
                  if(countFailures > 3) {
                     System.err.println("*** MAX FAILURES ENCOUNTERED - STOPPING THREAD " + Thread.currentThread().getName() + " ***");
                     setEnabled(false);
                     synchronized(example) {
                        example.notifyAll();
                     }
                  }
               }

               session.commit();
               consumer.close();

            }
            catch(final Exception e) {
               if(session != null) {
                  try {
                     session.rollback();
                  } catch(Exception e2) {}
               }
               setEnabled(false);
               e.printStackTrace();
               throw new RuntimeException(e);
            }
            finally {
               try {
                  if(session != null && !REUSE_SESSION) {
                     if(session.getTransacted()) {
                        session.commit();
                     }
                     session.close();
                  }
               } catch(Exception e) {
                  e.printStackTrace();
                  throw new RuntimeException(e);
               } finally {
                  returnConnection(conn);
               }
            }

            sleepRand();
         }
      }
   }

   @Override
   public boolean runExample() throws Exception
   {
      InitialContext initialContext = null;
      TestConnectionPool pool = null;
  
      try
      {
         // Step 1. Create an initial context to perform the JNDI lookup.
         initialContext = getContext(0);

         // Step 2. Look-up the JMS topic
         final Queue queue = (Queue)initialContext.lookup("/queue/exampleQueue");

         // Step 3. Look-up the JMS connection factory
         final ConnectionFactory cf = (ConnectionFactory)initialContext.lookup("/ConnectionFactory");
         pool = new TestConnectionPool(cf);

         final List<TestMessageConsumer> consumers = new ArrayList<TestMessageConsumer>(THREADS_CONSUMER);
         final List<Thread> consumerThreads = new ArrayList<Thread>(THREADS_CONSUMER);
         final List<TestMessageProducer> producers = new ArrayList<TestMessageProducer>(THREADS_PRODUCER);
         final List<Thread> producerThreads = new ArrayList<Thread>(THREADS_PRODUCER);

         for(int i=0; i < THREADS_CONSUMER; i++) {
            final TestMessageConsumer consumer = new TestMessageConsumer();
            final Thread consumerThread = new Thread(consumer);
            consumer.setConnectionPool(pool);
            consumer.setQueue(queue);
            consumer.setExample(this);
            consumers.add(consumer);
            consumerThreads.add(consumerThread);
         }
         for(int i=0; i < THREADS_PRODUCER; i++) {
            final TestMessageProducer producer = new TestMessageProducer();
            final Thread producerThread = new Thread(producer);
            producer.setConnectionPool(pool);
            producer.setQueue(queue);
            producers.add(producer);
            producerThreads.add(producerThread);
         }

         System.err.println("*** SINGLE CONNECTION = " + SINGLE_CONN + " ***");
         System.err.println("*** REUSE SESSIONS = " + REUSE_SESSION + " ***");

         for(final Thread producer : producerThreads) {
            producer.start();
         }
         for(final Thread consumer : consumerThreads) {
            consumer.start();
         }

         try {
            synchronized(this) {
               wait(TEST_TIME);
            }
         } catch(final InterruptedException ie) {

         }

         boolean stillRunning = true;
 
         System.out.println("Shutting down");
         for(final TestMessageProducer producer : producers) {
            producer.setEnabled(false);
         }
         for(final TestMessageConsumer consumer : consumers) {
            stillRunning &= consumer.isEnabled();
            consumer.setEnabled(false);
         }
         for(final Thread producer : producerThreads) {
            producer.join(1000l);
         }
         for(final Thread consumer : consumerThreads) {
            consumer.join(1000l);
         }

         System.out.println("*** Producer a total of " + TestMessageProducer.getCount() + " messages ***");
         System.out.println("*** Consumed a total of " + TestMessageConsumer.getCount() + " messages ***");

         return stillRunning;
      }
      finally
      {
         if (pool != null) 
         {
            pool.shutdown();
         }
         if (initialContext != null)
         {
            initialContext.close();
         }
      }
   }

}
